import json
import asyncio
import logging

from sagemaker_gen_ai_jupyterlab_extension.constants import Q_UNSUBSCRIBED_MESSAGE, INTERAL_SERVER_ERROR_MESSAGE
from .websocket_manager import manager
from tornado.websocket import WebSocketHandler
from uuid import uuid4
from typing import Optional

# Configure logging
logger = logging.getLogger('SageMakerGenAIJupyterLabExtension.websocket')

class WebSocketHandler(WebSocketHandler):
    def initialize(self):
        self.active_requests = {}
        self.active_tasks = set()
        self.chat_stream_tokens = {}
        
    def open(self):
        logger.info(f"WebSocket connection opened: {self.request.remote_ip}")
        manager.add_connection(self)

        from . import lsp_connection
        if lsp_connection:
            lsp_connection.set_send_to_client(manager.broadcast)
        else:
            logger.warning("LSP connection not available during WebSocket open")
    
    async def on_message(self, event):
        try:
            message = json.loads(event)
            message_id = message.get('id')
            command = message.get('command')
            logger.info(f"Received message: {command} (id: {message_id})")
            
            # Process message in background task to avoid blocking
            asyncio.create_task(self._process_message(message))
        except json.JSONDecodeError as e:
            logger.error(f"JSON parsing error: {e}")
            self.send_error_response(None, "unknown", f"Invalid JSON: {str(e)}")
        except Exception as e:
            logger.exception(f"Error processing message: {e}")
            # Extract ID if possible for error response
            try:
                message_dict = json.loads(event)
                message_id = message_dict.get('id')
                command = message_dict.get('command', 'unknown')
                self.send_error_response(message_id, command, str(e))
            except:
                logger.error("Could not extract message ID for error response")
                
    async def _process_message(self, message):
        """Process message in a separate task to avoid blocking the WebSocket handler"""
        task = asyncio.current_task()
        if task:
            self.active_tasks.add(task)
            
        try:
            await self.handle_inbound_event(message)
        except Exception as e:
            logger.exception(f"Error in background task: {e}")
            message_id = message.get('id')
            command = message.get('command', 'unknown')
            if message_id:
                self.send_error_response(message_id, command, str(e))
        finally:
            # Clean up task reference
            if task and task in self.active_tasks:
                self.active_tasks.remove(task)

    def send_error_response(self, message_id: Optional[str], command: str, error_message: str):
        """Send an error response for a failed request"""
        if message_id:
            error_response = {
                "id": message_id,
                "command": command,
                "error": error_message
            }
            manager.broadcast(json.dumps(error_response))
            logger.info(f"Sent error response for {command} (id: {message_id})")

    async def handle_inbound_event(self, message):
        from . import lsp_connection
        
        if not lsp_connection:
            logger.error("LSP connection not available")
            self.send_error_response(
                message.get('id'), 
                message.get('command', 'unknown'), 
                "LSP connection not available"
            )
            return
            
        id = message.get('id')
        command = message.get('command')
        params = message.get('params', {})
        tabId = message.get('tabId')
        
        logger.info(f"Processing command: {command}")

        try:
            match(command):
                case 'aws/chat/sendChatPrompt':
                    await self.handle_chat_prompt(lsp_connection, id, command, params, tabId)
                case 'aws/chat/buttonClick':
                    await self.handle_button_click(lsp_connection, id, command, params, tabId)
                case 'aws/chat/listMcpServers':
                    await self.handle_list_mcp_servers(lsp_connection, id, command, params, tabId)
                case 'aws/chat/mcpServerClick':
                    await self.handle_mcp_server_click(lsp_connection, id, command, params, tabId)
                case 'aws/chat/listConversations':
                    await self.list_conversations(lsp_connection, id, command, params, tabId)
                case 'aws/chat/conversationClick':
                    await self.conversation_click(lsp_connection, id, command, params, tabId)
                case 'aws/chat/ready':
                    # refresh credentials on each page load 
                    from . import get_credential_manager
                    credential_manager = get_credential_manager()
                    if credential_manager:
                        credential_manager.handle_tier_change()
                    
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/tabAdd':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/openTab':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/tabChange':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/tabRemove':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/listRules':
                    await self.list_rules(lsp_connection, id, command, params, tabId)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/pinnedContextAdd':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/pinnedContextRemove':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                case 'aws/chat/createPrompt':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/fileClick':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'errorMessage':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'stopChatResponse':
                    self.handle_stop_chat_response(lsp_connection, id, command, params)
                    logger.info(f"Sent notification: {command}")
                case 'workspace/didChangeConfiguration':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/promptInputOptionChange':
                    await asyncio.to_thread(lsp_connection.send_notification, command, params)
                    logger.debug(f"Sent notification: {command}")
                case 'aws/chat/sendChatQuickAction':
                    await self.handle_chat_quick_action(lsp_connection, id, command, params, tabId)
                case _:
                    logger.warning(f"Unhandled command: {command}")
                    if id:
                        self.send_error_response(id, command, f"Unsupported command: {command}")
        except Exception as e:
            logger.exception(f"Error handling command {command}: {e}")
            self.send_error_response(id, command, f"Error: {str(e)}")

    async def handle_chat_prompt(self, lsp_connection, id, command, params, tabId):
        """Handle chat prompt requests with streaming responses and cancellation support."""
        from .cancellation_token import CancellationTokenSource
        import time
        
        # User is sending a chat prompt, so we're no longer in startup mode
        if hasattr(lsp_connection, 'is_startup'):
            lsp_connection.is_startup = False
        
        if not tabId:
            self.send_error_response(id, command, "Missing tabId")
            return
            
        partial_result_token = str(uuid4())
        logger.debug(f"Starting chat prompt {id} with token: {partial_result_token}")
        
        # Setup cancellation tracking
        cancellation_token_source = CancellationTokenSource()
        last_partial_result = None
        
        self._store_chat_session(tabId, cancellation_token_source, lsp_connection.lsp_endpoint.next_id)
        
        def send_chat_update(partial_results):
            nonlocal last_partial_result
            if cancellation_token_source.token.isCancellationRequested:
                return
                
            last_partial_result = partial_results
            self._broadcast_partial_result(command, partial_results, tabId)

        dispose = None
        try:
            # Register progress handler
            dispose = await asyncio.to_thread(
                lsp_connection.on_progress, 
                partial_result_token, 
                send_chat_update,
                cancellation_token_source.token
            )
            
            # Make LSP call
            response = await asyncio.to_thread(
                lsp_connection.lsp_endpoint.call_method, 
                command, 
                **{**params, "partialResultToken": partial_result_token}
            )
            
            # Check for re-auth response and modify if needed
            if (response and isinstance(response, dict) and 
                response.get('followUp', {}).get('options')):
                for option in response['followUp']['options']:
                    if option.get('type') == 're-auth':
                        self.send_error_response(id, command, Q_UNSUBSCRIBED_MESSAGE)
                        break
            
            # Check for 500 error response
            if (response and isinstance(response, dict) and 
                response.get('error', {}).get('code') == 500):
                error_message = response.get('error', {}).get('message', 'Internal server error')
                logger.error(f"Server error in chat prompt {id}: {error_message}")
                self.send_error_response(id, command, INTERAL_SERVER_ERROR_MESSAGE)
                return

            # Handle response or cancellation
            self._send_success_response(id, command, tabId, response)

                
        except Exception as e:
            logger.exception(f"Error in chat prompt {id}: {e}")
            self.send_error_response(id, command, str(e))
        finally:
            self._cleanup_chat_session(tabId, dispose)
    
    def _store_chat_session(self, tab_id, token_source, lsp_request_id):
        """Store chat session for cancellation tracking."""
        self.chat_stream_tokens[tab_id] = {
            'token_source': token_source,
            'lsp_request_id': lsp_request_id
        }
        logger.debug(f"Stored session for tab {tab_id} with LSP ID {lsp_request_id}")
    
    def _broadcast_partial_result(self, command, partial_results, tab_id):
        """Broadcast partial result to client."""
        try:
            manager.broadcast({
                "command": command,
                "params": partial_results,
                "tabId": tab_id,
                "isPartialResult": True,
            })
        except Exception as e:
            logger.error(f"Error broadcasting partial result: {e}")
    
    def _send_success_response(self, id, command, tab_id, response):
        """Send successful chat response."""
        result = {
            "id": id,
            "command": command,
            "params": response,
            "tabId": tab_id
        }
        manager.broadcast(json.dumps(result))
        logger.info(f"Completed chat prompt: {id}")
    
    def _cleanup_chat_session(self, tab_id, dispose):
        """Clean up chat session resources."""
        if dispose:
            try:
                dispose()
            except Exception as e:
                logger.error(f"Error disposing progress handler: {e}")
        
        if tab_id in self.chat_stream_tokens:
            del self.chat_stream_tokens[tab_id]

    async def handle_button_click(self, lsp_connection, id, command, params, tabId):
        """Handle button click requests"""
        logger.info(f"Processing button click: {params.get('buttonId', 'unknown')}")
        
        try:
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
            
            # Send success response
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            manager.broadcast(json.dumps(result))
            logger.info(f"Button click processed successfully")
        except Exception as e:
            logger.exception(f"Button click error: {e}")
            self.send_error_response(id, command, str(e))

    async def handle_list_mcp_servers(self, lsp_connection, id, command, params, tabId):
        """Handle list MCP servers requests"""
        logger.info(f"Listing MCP servers")
        
        try:
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
            
            # Send success response
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            manager.broadcast(json.dumps(result))
            logger.info(f"Listed MCP servers successfully")
        except Exception as e:
            logger.exception(f"List MCP servers error: {e}")
            self.send_error_response(id, command, str(e))

    async def handle_chat_quick_action(self, lsp_connection, id, command, params, tabId):
        """Handle chat quick actions"""
        logger.info(f"Handling chat quick action")
        
        try:
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
 
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            
            manager.broadcast(json.dumps(result))
            logger.info(f"Handled Chat Quick Action successfully")
        except Exception as e:
            logger.exception(f"Handle Chat Quick Action error: {e}")
            self.send_error_response(id, command, str(e))
 
    async def handle_mcp_server_click(self, lsp_connection, id, command, params, tabId):
        """Handle MCP server click requests"""
        logger.info(f"Processing MCP server click")
        
        try:
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
            
            # Send success response
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            manager.broadcast(json.dumps(result))
            logger.info(f"MCP server click processed successfully")
        except Exception as e:
            logger.exception(f"MCP server click error: {e}")
            self.send_error_response(id, command, str(e))
    
    async def list_conversations(self, lsp_connection, id, command, params, tabId):
        """Handle list conversation requests"""
        logger.info(f"Listing conversations")
        
        try:
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
            
            # Send success response
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            manager.broadcast(json.dumps(result))
            logger.info(f"Listed conversations successfully")
        except Exception as e:
            logger.exception(f"List conversations error: {e}")
            self.send_error_response(id, command, str(e))

    async def list_rules(self, lsp_connection, id, command, params, tabId):
        """Handle list rules request"""
        logger.info(f"Listing rules")
        
        try:
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
            
            # Send success response
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            manager.broadcast(json.dumps(result))
            logger.info(f"Listed rules successfully")
        except Exception as e:
            logger.exception(f"List rules error: {e}")
            self.send_error_response(id, command, str(e))

    async def conversation_click(self, lsp_connection, id, command, params, tabId):
        """Handle conversation click requests"""
        logger.info(f"Processing conversation click")
        
        try:
            # User clicked on a conversation, so we're no longer in startup mode
            if hasattr(lsp_connection, 'is_startup'):
                lsp_connection.is_startup = False
                
            response = await asyncio.to_thread(lsp_connection.call_method, command, params)
            
            # Send success response
            result = {
                "id": id,
                "command": command,
                "params": response,
                "tabId": tabId
            }
            manager.broadcast(json.dumps(result))
            logger.info(f"Conversation click processed successfully")
        except Exception as e:
            logger.exception(f"Conversation click error: {e}")
            self.send_error_response(id, command, str(e))

    def on_close(self):
        logger.info(f"WebSocket connection closed: {self.request.remote_ip}")
        
        # Cancel any running tasks when connection closes
        for task in list(self.active_tasks):
            if not task.done():
                logger.info(f"Cancelling task {task}")
                task.cancel()
        
        manager.remove_connection(self)

    def check_origin(self, origin):
        # Only allow connections from the same origin as the server
        server_origin = self.request.host
        client_origin = origin.replace("http://", "").replace("https://", "")
        logger.info(f"Origin check - Client origin: {client_origin}, Server origin: {server_origin}")
        return client_origin == server_origin
    
    def handle_stop_chat_response(self, lsp_connection, id, command, params):
        tab_id = params.get('tabId')
        logger.debug(f'Cancelling chat response for {tab_id}')
        if tab_id in self.chat_stream_tokens:
            token_info = self.chat_stream_tokens[tab_id]
            token_source = token_info['token_source']
            
            # Cancel token
            token_source.cancel()
            # Send $/cancelRequest to LSP server
            lsp_request_id = token_info.get('lsp_request_id')
            if lsp_request_id:
                try:
                    asyncio.create_task(
                        asyncio.to_thread(
                            lsp_connection.send_notification,
                            "$/cancelRequest",
                            {"id": lsp_request_id}
                        )
                    )
                    logger.info(f"Sent $/cancelRequest for LSP request: {lsp_request_id}")
                except Exception as e:
                    logger.error(f"Error sending $/cancelRequest: {e}")
            
            # Remove from tracking
            del self.chat_stream_tokens[tab_id]
            logger.info(f"Cancelled chat for tab: {tab_id}")



class CancellationToken:
    """Token that can be used to signal cancellation of an operation."""
    def __init__(self):
        self._is_cancelled = False
    
    @property
    def isCancellationRequested(self) -> bool:
        return self._is_cancelled
    
    def cancel(self):
        self._is_cancelled = True

class CancellationTokenSource:
    """Source for creating and controlling cancellation tokens."""
    def __init__(self):
        self._token = CancellationToken()
    
    @property
    def token(self) -> CancellationToken:
        return self._token
    
    def cancel(self):
        self._token.cancel()


import unittest
import tempfile
import os
import json
import time
import sys
from unittest.mock import Mock, patch, MagicMock

# Mock watchdog before importing
with patch.dict('sys.modules', {
    'watchdog': Mock(),
    'watchdog.observers': Mock(),
    'watchdog.events': Mock()
}):
    from ..q_customization import QCustomization


class TestQCustomization(unittest.TestCase):
    
    def setUp(self):
        self.mock_lsp_connection = Mock()
        self.watcher = QCustomization(self.mock_lsp_connection)
        
    def tearDown(self):
        self.watcher.stop_watcher_for_customization_file()
    
    def test_start_watcher(self):
        with patch.object(self.watcher.file_watcher, 'start_watching') as mock_start:
            self.watcher.start_customization_file_watcher()
            mock_start.assert_called_once()
        
    def test_stop_watcher(self):
        with patch.object(self.watcher.file_watcher, 'stop_watching') as mock_stop:
            self.watcher.stop_watcher_for_customization_file()
            mock_stop.assert_called_once()
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.q_customization.extract_q_customization_arn')
    def test_handle_file_change(self, mock_extract):
        mock_extract.return_value = 'test-arn'
        
        self.watcher._handle_customization_file_change()
        
        self.mock_lsp_connection.update_q_customization.assert_called_once_with('test-arn')


if __name__ == '__main__':
    unittest.main()


import logging
import threading
import boto3
from .extract_utils import extract_bearer_token, extract_auth_mode, AuthMode, extract_q_settings

logger = logging.getLogger('SageMakerGenAIJupyterLabExtension')
START_URL = "randomstring"
REFRESH_TIMER_LIMIT = 300
CREDS_PROFILE_NAME = "DomainExecutionRoleCreds"

class CredentialManager:
    def __init__(self, lsp_connection, auth_mode):
        self.lsp_connection = lsp_connection
        self.current_auth_mode = auth_mode
        self.refresh_timer = None
        
    def initialize_credentials(self):
        """Initialize credentials based on current auth mode"""
        if self.current_auth_mode == AuthMode.IDC.value:
            self._update_bearer_token()
            self._start_refresh(self.current_auth_mode)
            logger.info("Configured bearer token authentication")
        elif self.current_auth_mode == AuthMode.IAM.value:
            self._update_iam_credentials()
            self._start_refresh(self.current_auth_mode)
            logger.info("Configured IAM SigV4 authentication")
        else:
            logger.error(f"Invalid auth mode: {self.current_auth_mode}")
    
    def handle_tier_change(self):
        """Handle Q tier changes by switching credential types"""
        new_auth_mode = extract_auth_mode()
        
        if new_auth_mode != self.current_auth_mode:
            logger.info(f"Q tier changed: {self.current_auth_mode} -> {new_auth_mode}")
            self._stop_refresh_timer()
            
            # Restart LSP server with new auth mode
            from . import NODE_PATH, LSP_EXECUTABLE_PATH, WORKSPACE_FOLDER
            self.lsp_connection.restart_lsp_server(NODE_PATH, LSP_EXECUTABLE_PATH, new_auth_mode, WORKSPACE_FOLDER)
            
            if new_auth_mode == AuthMode.IDC.value:
                # Free → Pro: Switch to bearer token
                self._update_bearer_token()
                self._start_refresh(new_auth_mode)
                self._setup_paid_tier_features()
                logger.info("Configured bearer token authentication")
            elif new_auth_mode == AuthMode.IAM.value:
                # Pro → Free: Switch to IAM SigV4
                self._update_iam_credentials()
                self._start_refresh(new_auth_mode)
                logger.info("Configured IAM SigV4 authentication")
            
            self.current_auth_mode = new_auth_mode
        else:
            logger.info(f"Q tier unchanged: {self.current_auth_mode}")
            self._stop_refresh_timer()
            self.initialize_credentials()
    

    def _update_iam_credentials(self):
        """Update IAM credentials to LSP server"""
        try:
            credentials = self._get_iam_credentials()
            if credentials:
                self.lsp_connection.update_iam_credentials(credentials.access_key, credentials.secret_key, credentials.token, str(credentials._expiry_time))
                logger.info("Successfully updated IAM credentials")
            else:
                logger.error("Failed to setup IAM credentials")
        except Exception as e:
            logger.error(f"Error refreshing IAM credentials: {e}")

    def _get_iam_credentials(self):
        try:
            session = boto3.Session(profile_name=CREDS_PROFILE_NAME)
            credentials = session.get_credentials()
            return credentials
        except boto3.exceptions.Boto3Error as boto_error:
            logger.error(f"AWS credentials error: {str(boto_error)}")
        except Exception as e:
            logger.error(f"Failed to get credentials: {str(e)}")

    
    def _update_bearer_token(self):
        """Update bearer token to LSP server"""
        try:
            token = extract_bearer_token()
            # We will set start URL to Q profile ARN so it can be used 
            # to distunguish customer requests for telemetry
            start_url = extract_q_settings()
            if start_url is None:
                # Setting it to a randomstring if there is error getting q profile ARN
                # since Flare does not accept None start URL
                # Ref: https://t.corp.amazon.com/P250511444/communication
                start_url = START_URL
            if token:
                self.lsp_connection.update_access_token(token, start_url)
            else:
                logger.error("Failed to refresh bearer token")
        except Exception as e:
            logger.error(f"Error refreshing bearer token: {e}")
    
    def _start_refresh(self, auth_mode):
        """Start periodic bearer token/IAM credentials refresh"""
        def refresh_and_reschedule():
            if auth_mode == AuthMode.IDC.value:
                self._update_bearer_token()
            elif auth_mode == AuthMode.IAM.value:
                self._update_iam_credentials()
            self._start_refresh(auth_mode)
        # setting the next refresh
        self.refresh_timer = threading.Timer(REFRESH_TIMER_LIMIT, refresh_and_reschedule)
        self.refresh_timer.daemon = True
        self.refresh_timer.start()
    
    def _stop_refresh_timer(self):
        """Stop the current refresh timer"""
        if self.refresh_timer:
            self.refresh_timer.cancel()
            self.refresh_timer = None
    
    def _setup_paid_tier_features(self):
        """Setup Q profile and customization for paid tier"""
        try:
            from .extract_utils import extract_q_settings, extract_q_customization_arn
            
            # Update Q profile
            q_settings = extract_q_settings()
            if q_settings:
                self.lsp_connection.update_q_profile(q_settings)
                logger.info("Updated Q profile for paid tier")
            
            # Update Q customization
            customization_arn = extract_q_customization_arn()
            if customization_arn:
                self.lsp_connection.update_q_customization(customization_arn)
                logger.info("Updated Q customization for paid tier")
                
        except Exception as e:
            logger.error(f"Error setting up paid tier features: {e}")
    
    def cleanup(self):
        """Cleanup resources"""
        self._stop_refresh_timer()


from unittest.mock import Mock, patch, mock_open
import json
import pytest
import os

from sagemaker_gen_ai_jupyterlab_extension.extract_utils import (
    extract_bearer_token, extract_q_settings, extract_auth_mode, 
    extract_q_customization_arn, AuthMode, FilePaths
)

class TestExtractUtils:
    
    def test_auth_mode_enum(self):
        """Test AuthMode enum values"""
        assert AuthMode.IDC.value == "IDC"
        assert AuthMode.IAM.value == "IAM"
    
    @patch('builtins.open', new_callable=mock_open, read_data='{"idc_access_token": "test_token"}')
    @patch('os.path.expanduser')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_bearer_token_success(self, mock_logger, mock_expanduser, mock_file):
        """Test successful bearer token extraction"""
        mock_expanduser.return_value = "/home/user/.aws/sso/idc_access_token.json"
        
        token = extract_bearer_token()
        
        assert token == "test_token"
        mock_expanduser.assert_called_once_with("~/.aws/sso/idc_access_token.json")

    @patch('builtins.open', side_effect=FileNotFoundError)
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_bearer_token_file_not_found(self, mock_logger, mock_file):
        """Test bearer token extraction when file not found"""
        token = extract_bearer_token()
        
        assert token is None
        mock_logger.error.assert_called_with(f"Token file not found: {FilePaths.BEARER_TOKEN.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='invalid json')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_bearer_token_invalid_json(self, mock_logger, mock_file):
        """Test bearer token extraction with invalid JSON"""
        token = extract_bearer_token()
        
        assert token is None
        mock_logger.error.assert_called_with(f"Invalid JSON in token file: {FilePaths.BEARER_TOKEN.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='{"idc_access_token": ""}')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_bearer_token_empty_token(self, mock_logger, mock_file):
        """Test bearer token extraction with empty token"""
        token = extract_bearer_token()
        
        assert token is None
        mock_logger.error.assert_called_with(f"No bearer token found in {FilePaths.BEARER_TOKEN.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='{"wrong_key": "test_token"}')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_bearer_token_wrong_key(self, mock_logger, mock_file):
        """Test bearer token extraction with wrong key"""
        token = extract_bearer_token()
        
        assert token is None
        mock_logger.error.assert_called_with(f"No bearer token found in {FilePaths.BEARER_TOKEN.value}")

    @patch('builtins.open', new_callable=mock_open, read_data='{"q_dev_profile_arn": "test_arn"}')
    @patch('os.path.expanduser')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_settings_success(self, mock_logger, mock_expanduser, mock_file):
        """Test successful Q settings extraction"""
        mock_expanduser.return_value = "/home/user/.aws/amazon_q/q_dev_profile.json"
        
        arn = extract_q_settings()
        
        assert arn == "test_arn"
        mock_logger.info.assert_called_with("Returning Q profile ARN: test_arn")
        
    @patch('builtins.open', side_effect=FileNotFoundError)
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_settings_file_not_found(self, mock_logger, mock_file):
        """Test Q settings extraction when file not found"""
        arn = extract_q_settings()
        
        assert arn is None
        mock_logger.error.assert_called_with(f"Profile  not found: {FilePaths.Q_DEV_PROFILE.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='invalid json')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_settings_invalid_json(self, mock_logger, mock_file):
        """Test Q settings extraction with invalid JSON"""
        arn = extract_q_settings()
        
        assert arn is None
        mock_logger.error.assert_called_with(f"Invalid JSON in token file: {FilePaths.Q_DEV_PROFILE.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='{"q_dev_profile_arn": ""}')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_settings_empty_arn(self, mock_logger, mock_file):
        """Test Q settings extraction with empty ARN"""
        arn = extract_q_settings()
        
        assert arn is None
        mock_logger.error.assert_called_with(f"No Q profile ARN found in {FilePaths.Q_DEV_PROFILE.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='{"wrong_key": "test_arn"}')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_settings_wrong_key(self, mock_logger, mock_file):
        """Test Q settings extraction with wrong key"""
        arn = extract_q_settings()
        
        assert arn is None
        mock_logger.error.assert_called_with(f"No Q profile ARN found in {FilePaths.Q_DEV_PROFILE.value}")

    @patch('builtins.open', new_callable=mock_open, read_data='{"auth_mode": "IDC"}')
    @patch('os.path.expanduser')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_idc(self, mock_logger, mock_expanduser, mock_file):
        """Test auth mode extraction with IDC value"""
        mock_expanduser.return_value = "/home/user/.aws/amazon_q/settings.json"
        
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IDC.value
        mock_logger.info.assert_called_with("Returning auth mode: IDC")

    @patch('builtins.open', new_callable=mock_open, read_data='{"auth_mode": "IAM"}')
    @patch('os.path.expanduser')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_iam(self, mock_logger, mock_expanduser, mock_file):
        """Test auth mode extraction with IAM value"""
        mock_expanduser.return_value = "/home/user/.aws/amazon_q/settings.json"
        
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IAM.value
        mock_logger.info.assert_called_with("Returning auth mode: IAM")

    @patch('builtins.open', new_callable=mock_open, read_data='{"auth_mode": "INVALID"}')
    @patch('os.path.expanduser')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_invalid(self, mock_logger, mock_expanduser, mock_file):
        """Test auth mode extraction with invalid value"""
        mock_expanduser.return_value = "/home/user/.aws/amazon_q/settings.json"
        
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IAM.value
        mock_logger.error.assert_called_with("Invalid auth mode: INVALID, defaulting to IAM")
        
    @patch('builtins.open', new_callable=mock_open, read_data='{"auth_mode": ""}')
    @patch('os.path.expanduser')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_empty(self, mock_logger, mock_expanduser, mock_file):
        """Test auth mode extraction with empty value"""
        mock_expanduser.return_value = "/home/user/.aws/amazon_q/settings.json"
        
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IAM.value
        mock_logger.error.assert_called_with(f"No auth mode found in {FilePaths.SETTINGS.value}")
        
    @patch('builtins.open', side_effect=FileNotFoundError)
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_file_not_found(self, mock_logger, mock_file):
        """Test auth mode extraction when file not found"""
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IAM.value
        mock_logger.error.assert_called_with(f"Profile not found: {FilePaths.SETTINGS.value}")
        
    @patch('builtins.open', new_callable=mock_open, read_data='invalid json')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_invalid_json(self, mock_logger, mock_file):
        """Test auth mode extraction with invalid JSON"""
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IAM.value
        mock_logger.error.assert_called_with(f"Invalid JSON in token file: {FilePaths.SETTINGS.value}")
        
    @patch('builtins.open')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_auth_mode_general_exception(self, mock_logger, mock_file):
        """Test auth mode extraction with general exception"""
        mock_file.side_effect = Exception("Test error")
        
        auth_mode = extract_auth_mode()
        
        assert auth_mode == AuthMode.IAM.value
        mock_logger.error.assert_called_with(f"Error reading token file: Test error")



    @patch('os.path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='{"customization_arn": "test_arn"}')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_customization_arn_success(self, mock_logger, mock_file, mock_exists):
        """Test successful customization ARN extraction"""
        mock_exists.return_value = True
        
        arn = extract_q_customization_arn()
        
        assert arn == "test_arn"  # The function looks for 'customization_arn' key
        mock_logger.info.assert_any_call("Found Q customization ARN: test_arn")
        
    @patch('os.path.exists')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_customization_arn_file_not_exists(self, mock_logger, mock_exists):
        """Test customization ARN extraction when file doesn't exist"""
        mock_exists.return_value = False
        
        arn = extract_q_customization_arn()
        
        assert arn is None
        mock_logger.info.assert_any_call(f"Customization file does not exist at: {os.path.expanduser(FilePaths.Q_CUSTOMIZATION_ARN.value)}")
        
    @patch('os.path.exists')
    @patch('builtins.open', side_effect=FileNotFoundError)
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_customization_arn_file_not_found(self, mock_logger, mock_file, mock_exists):
        """Test customization ARN extraction when file not found"""
        mock_exists.return_value = True
        
        arn = extract_q_customization_arn()
        
        assert arn is None
        mock_logger.info.assert_called_with(f"Customization file not found: {os.path.expanduser(FilePaths.Q_CUSTOMIZATION_ARN.value)}")
        
    @patch('os.path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='invalid json')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_customization_arn_invalid_json(self, mock_logger, mock_file, mock_exists):
        """Test customization ARN extraction with invalid JSON"""
        mock_exists.return_value = True
        
        arn = extract_q_customization_arn()
        
        assert arn is None
        mock_logger.error.assert_called_with(f"Invalid JSON in customization file: {os.path.expanduser(FilePaths.Q_CUSTOMIZATION_ARN.value)}")
        
    @patch('os.path.exists')
    @patch('builtins.open', new_callable=mock_open, read_data='{"customization_arn": ""}')
    @patch('sagemaker_gen_ai_jupyterlab_extension.extract_utils.logger')
    def test_extract_q_customization_arn_empty(self, mock_logger, mock_file, mock_exists):
        """Test customization ARN extraction with empty value"""
        mock_exists.return_value = True
        
        arn = extract_q_customization_arn()
        
        assert arn is None
        mock_logger.info.assert_any_call(f"No customization ARN found in {os.path.expanduser(FilePaths.Q_CUSTOMIZATION_ARN.value)}")


import unittest
from unittest.mock import patch, MagicMock
import os
from sagemaker_gen_ai_jupyterlab_extension.telemetry_collector import TelemetryCollector


class TestTelemetryCollector(unittest.TestCase):
    
    def test_extract_account_id_from_arn_valid(self):
        """Test extracting account ID from valid ARN"""
        arn = "arn:aws:codewhisperer:us-east-1:756160874543:profile/N4X4X9CQEXRP"
        account_id = TelemetryCollector._extract_account_id_from_arn(arn)
        self.assertEqual(account_id, "756160874543")
    
    def test_extract_account_id_from_arn_invalid(self):
        """Test extracting account ID from invalid ARN"""
        invalid_arn = "invalid-arn-format"
        account_id = TelemetryCollector._extract_account_id_from_arn(invalid_arn)
        self.assertIsNone(account_id)
    
    def test_extract_account_id_from_arn_none(self):
        """Test extracting account ID from None"""
        account_id = TelemetryCollector._extract_account_id_from_arn(None)
        self.assertIsNone(account_id)
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.EnvironmentDetector.get_environment')
    @patch('asyncio.new_event_loop')
    def test_detect_environment_md_idc(self, mock_new_loop, mock_get_env):
        """Test environment detection for MD_IDC"""
        mock_env = MagicMock()
        mock_env.name = "MD_IDC"
        mock_get_env.return_value = mock_env
        
        mock_loop = MagicMock()
        mock_new_loop.return_value = mock_loop
        mock_loop.run_until_complete.return_value = mock_env
        
        env = TelemetryCollector._detect_environment()
        self.assertEqual(env, "MD_IDC")
    
    @patch('asyncio.run')
    def test_detect_environment_exception(self, mock_run):
        """Test environment detection fallback on exception"""
        mock_run.side_effect = Exception("Test error")
        
        env = TelemetryCollector._detect_environment()
        self.assertEqual(env, "UNKNOWN")
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.TelemetryCollector._detect_environment')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.get_new_metrics_context')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.flush_metrics')
    def test_collect_telemetry_basic(self, mock_flush, mock_get_context, mock_detect_env):
        """Test basic telemetry collection"""
        mock_detect_env.return_value = "MD_IDC"
        mock_metrics = MagicMock()
        mock_get_context.return_value = mock_metrics
        
        telemetry_params = {
            "name": "amazonq_enterFocusConversation",
            "data": {
                "latency": 150,
                "result": "Succeeded"
            }
        }
        
        TelemetryCollector.collect_telemetry(telemetry_params)
        
        # Verify metrics context was created
        mock_get_context.assert_called_once_with("LSP_Telemetry_Event", "MD_IDC")
        
        # Verify dimensions were added (EventName and Result)
        expected_dimension_calls = [
            unittest.mock.call({"EventName": "amazonq_enterFocusConversation"}),
            unittest.mock.call({"Result": "Succeeded"})
        ]
        mock_metrics.put_dimensions.assert_has_calls(expected_dimension_calls, any_order=True)
        
        # Verify metric was added
        mock_metrics.put_metric.assert_called_with("RequestLatency", 150, "Milliseconds")
        
        # Verify metrics were flushed
        mock_flush.assert_called_once_with(mock_metrics)
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.TelemetryCollector._detect_environment')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.get_new_metrics_context')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.flush_metrics')
    def test_collect_telemetry_with_account_id(self, mock_flush, mock_get_context, mock_detect_env):
        """Test telemetry collection with account ID extraction"""
        mock_detect_env.return_value = "MD_IDC"
        mock_metrics = MagicMock()
        mock_get_context.return_value = mock_metrics
        
        telemetry_params = {
            "name": "test_event",
            "data": {
                "credentialStartUrl": "arn:aws:codewhisperer:us-east-1:756160874543:profile/test",
                "cwsprChatRequestLength": 100
            }
        }
        
        TelemetryCollector.collect_telemetry(telemetry_params)
        
        # Verify account ID was set (check that it was called at some point)
        mock_metrics.set_property.assert_any_call("AccountId", "756160874543")
        
        # Verify QProfileArn was also set
        mock_metrics.set_property.assert_any_call("QProfileArn", "arn:aws:codewhisperer:us-east-1:756160874543:profile/test")
        
        # Verify metric was added with correct mapping
        mock_metrics.put_metric.assert_called_with("ChatRequestSize", 100, "Count")
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.TelemetryCollector._detect_environment')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.get_new_metrics_context')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.flush_metrics')
    def test_collect_telemetry_string_to_int_conversion(self, mock_flush, mock_get_context, mock_detect_env):
        """Test string to int conversion in telemetry collection"""
        mock_detect_env.return_value = "MD_IDC"
        mock_metrics = MagicMock()
        mock_get_context.return_value = mock_metrics
        
        telemetry_params = {
            "name": "test_event",
            "data": {
                "latency": "250"  # String number
            }
        }
        
        TelemetryCollector.collect_telemetry(telemetry_params)
        
        # Verify string was converted to int
        mock_metrics.put_metric.assert_called_with("RequestLatency", 250, "Milliseconds")
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.TelemetryCollector._detect_environment')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.get_new_metrics_context')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.flush_metrics')
    def test_collect_telemetry_with_common_properties(self, mock_flush, mock_get_context, mock_detect_env):
        """Test telemetry collection with common properties"""
        mock_detect_env.return_value = "MD_IDC"
        mock_metrics = MagicMock()
        mock_get_context.return_value = mock_metrics
        
        telemetry_params = {
            "name": "test_event",
            "data": {
                "cwsprChatConversationId": "conv-123",
                "cwsprChatMessageId": "msg-456",
                "result": "Succeeded",
                "cwsprChatConversationType": "chat"
            }
        }
        
        TelemetryCollector.collect_telemetry(telemetry_params)
        
        # Verify properties were set
        expected_property_calls = [
            unittest.mock.call("ConversationID", "conv-123"),
            unittest.mock.call("MessageID", "msg-456")
        ]
        mock_metrics.set_property.assert_has_calls(expected_property_calls, any_order=True)
        
        # Verify dimensions were set
        expected_dimension_calls = [
            unittest.mock.call({"EventName": "test_event"}),
            unittest.mock.call({"Result": "Succeeded"}),
            unittest.mock.call({"ConversationType": "chat"})
        ]
        mock_metrics.put_dimensions.assert_has_calls(expected_dimension_calls, any_order=True)
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.logger')
    @patch('sagemaker_gen_ai_jupyterlab_extension.telemetry_collector.TelemetryCollector._detect_environment')
    def test_collect_telemetry_exception_handling(self, mock_detect_env, mock_logger):
        """Test exception handling in telemetry collection"""
        mock_detect_env.side_effect = Exception("Test error")
        
        telemetry_params = {"name": "test_event", "data": {}}
        
        TelemetryCollector.collect_telemetry(telemetry_params)
        
        # Verify error was logged
        mock_logger.error.assert_called_once_with("Error collecting LSP telemetry: Test error")





if __name__ == '__main__':
    unittest.main()


# Original Copyright https://github.com/yeger00/pylspclient. Licensed under MIT License.
from typing import Any, Optional
from enum import IntEnum


class ErrorCodes(IntEnum):
    # Defined by JSON RPC
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    serverErrorStart = -32099
    serverErrorEnd = -32000
    ServerNotInitialized = -32002
    UnknownErrorCode = -32001

    # Defined by the protocol.
    RequestCancelled = -32800
    ContentModified = -32801


class ResponseError(Exception):
    def __init__(self, code: ErrorCodes, message: str, data: Optional[Any] = None):
        self.code = code
        self.message = message
        if data:
            self.data = data



import pytest
import json
import os
from unittest.mock import Mock, patch

from sagemaker_gen_ai_jupyterlab_extension.handlers import RouteHandler, setup_handlers


class TestRouteHandler:
    
    @pytest.fixture
    def handler(self):
        """Create a RouteHandler instance for testing"""
        handler = Mock(spec=RouteHandler)
        handler.finish = Mock()
        return handler
    
    def test_get_endpoint(self):
        """Test the GET endpoint returns correct JSON response"""
        handler = Mock()
        handler.finish = Mock()
        
        # Call the actual method
        RouteHandler.get(handler)
        
        expected_response = json.dumps({
            "data": "This is /sagemaker_gen_ai_jupyterlab_extension/get-example endpoint!"
        })
        handler.finish.assert_called_once_with(expected_response)


class TestSetupHandlers:
    
    @pytest.fixture
    def mock_web_app(self):
        """Create a mock web app for testing"""
        web_app = Mock()
        web_app.settings = {"base_url": "/"}
        web_app.add_handlers = Mock()
        return web_app
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.manager')
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.ioloop.IOLoop')
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.logger')
    def test_setup_handlers_success(self, mock_logger, mock_ioloop, mock_manager, mock_web_app):
        """Test successful handler setup"""
        mock_current_loop = Mock()
        mock_ioloop.current.return_value = mock_current_loop
        
        setup_handlers(mock_web_app)
        
        # Verify handlers were added
        mock_web_app.add_handlers.assert_called_once()
        
        # Verify IOLoop was set on manager
        mock_manager.set_io_loop.assert_called_once_with(mock_current_loop)
        
        # Check that 3 handlers were registered (route, websocket, static)
        call_args = mock_web_app.add_handlers.call_args
        host_pattern, handlers = call_args[0]
        
        assert host_pattern == ".*$"
        assert len(handlers) == 3
        
        # Verify handler patterns
        route_pattern, route_handler = handlers[0]
        websocket_pattern, websocket_handler = handlers[1]
        static_pattern, static_handler, static_config = handlers[2]
        
        assert "/sagemaker_gen_ai_jupyterlab_extension/get-example" in route_pattern
        assert "/sagemaker_gen_ai_jupyterlab_extension/ws" in websocket_pattern
        assert "/sagemaker_gen_ai_jupyterlab_extension/static" in static_pattern
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.manager')
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.ioloop.IOLoop')
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.logger')
    def test_setup_handlers_exception(self, mock_logger, mock_ioloop, mock_manager, mock_web_app):
        """Test handler setup with exception"""
        mock_web_app.add_handlers.side_effect = Exception("Test error")
        
        setup_handlers(mock_web_app)
        
        # Verify error was logged
        mock_logger.error.assert_called_once_with("Error setting up handlers: Test error")
    
    def test_cors_headers_set(self, mock_web_app):
        """Test that CORS headers are properly set on RouteHandler"""
        with patch('sagemaker_gen_ai_jupyterlab_extension.handlers.manager'), \
             patch('sagemaker_gen_ai_jupyterlab_extension.handlers.ioloop.IOLoop'):
            
            setup_handlers(mock_web_app)
            
            # Test that CORS methods were added to RouteHandler
            assert hasattr(RouteHandler, 'set_default_headers')
            assert hasattr(RouteHandler, 'options')
            
            # Test CORS headers method
            handler = Mock()
            handler.set_header = Mock()
            
            RouteHandler.set_default_headers(handler)
            
            # Verify CORS headers are set
            handler.set_header.assert_any_call("Access-Control-Allow-Origin", "*")
            handler.set_header.assert_any_call("Access-Control-Allow-Headers", "x-requested-with, content-type, authorization")
            handler.set_header.assert_any_call("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
    
    def test_options_method(self, mock_web_app):
        """Test OPTIONS method for CORS preflight"""
        with patch('sagemaker_gen_ai_jupyterlab_extension.handlers.manager'), \
             patch('sagemaker_gen_ai_jupyterlab_extension.handlers.ioloop.IOLoop'):
            
            setup_handlers(mock_web_app)
            
            # Test OPTIONS method
            handler = Mock()
            handler.set_status = Mock()
            handler.finish = Mock()
            
            RouteHandler.options(handler)
            
            handler.set_status.assert_called_once_with(204)
            handler.finish.assert_called_once()
    
    @patch('sagemaker_gen_ai_jupyterlab_extension.handlers.CLIENT_HTML_PATH')
    def test_client_html_path_constant(self, mock_path):
        """Test that CLIENT_HTML_PATH is properly set"""
        expected_path = os.path.join(os.path.dirname(__file__), "static", "client.html")
        # The actual path should be set during import, so we just verify the structure
        assert "static" in str(expected_path)
        assert "client.html" in str(expected_path)


"""
SMUS Flare JupyterLab extension for AWS Q integration
"""
try:
    from ._version import __version__
except ImportError:
    # Fallback when using the package in dev mode without installing
    # in editable mode with pip. It is highly recommended to install
    # the package from a stable release or in editable mode: https://pip.pypa.io/en/stable/topics/local-project-installs/#editable-installs
    import warnings
    warnings.warn("Importing 'SageMakerGenAIJupyterLabExtension' outside a proper installation.")
    __version__ = "dev"
from .handlers import setup_handlers
from .request_logger import init_api_operation_logger
import logging
import os
import pathlib
import requests
import zipfile
import tempfile
from .lsp_server_connection import LspServerConnection
from .extract_utils import extract_q_customization_arn, extract_q_settings, extract_auth_mode
from .q_customization import QCustomization
from .credential_manager import CredentialManager

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('SageMakerGenAIJupyterLabExtension')

# Get the directory where this __init__.py file is located
PACKAGE_DIR = pathlib.Path(__file__).parent.absolute()
MANIFEST_URL = "https://aws-toolkit-language-servers.amazonaws.com/qAgenticChatServer/0/manifest.json"
FLARE_SERVER_VERSION = "1.23.0"
NODE_PATH = "/opt/conda/bin/node"
WORKSPACE_FOLDER = "file:///home/sagemaker-user/src"

# Global variable to store the LSP executable path
LSP_EXECUTABLE_PATH = None

# Global variable to store the LSP connection
lsp_connection = None

# Global variable to store the credential manager
credential_manager = None

def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": "sagemaker_gen_ai_jupyterlab_extension"
    }]

__version__ = '1.0.1'

def get_lsp_connection():
    """
    Returns the initialized LSP connection.
    """
    global lsp_connection
    return lsp_connection

def get_credential_manager():
    """
    Returns the initialized credential manager.
    """
    global credential_manager
    return credential_manager

def chat_with_prompt(prompt):
    """
    Frontend-facing function that only requires a prompt.
    Uses the pre-initialized LSP connection from __init__.py.
    
    Args:
        prompt (str): The chat prompt from the user
        
    Returns:
        dict: The response from the LSP server
    """
    lsp_connection = get_lsp_connection()
    return lsp_connection.get_chat_response(prompt)

def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": "sagemaker_gen_ai_jupyterlab_extension"
    }]

q_customization = None

def download_and_extract_lsp_server():
    """Download manifest, find Linux version, download and extract LSP server"""
    global LSP_EXECUTABLE_PATH
    
    try:
        # Download manifest
        logger.info(f"Downloading manifest from {MANIFEST_URL}")
        response = requests.get(MANIFEST_URL)
        response.raise_for_status()
        manifest = response.json()
        
        # Find the specified version
        version_data = None
        for version in manifest.get('versions', []):
            if version.get('serverVersion') == FLARE_SERVER_VERSION:
                version_data = version
                break
        
        if not version_data:
            raise Exception(f"Version {FLARE_SERVER_VERSION} not found in manifest")
        
        # Find Linux target
        linux_target = None
        for target in version_data.get('targets', []):
            if target.get('platform') == 'linux':
                linux_target = target
                break
        
        if not linux_target:
            raise Exception("Linux target not found in manifest")
        
        # Find servers.zip
        servers_zip_url = None
        for content in linux_target.get('contents', []):
            if content.get('filename') == 'servers.zip':
                servers_zip_url = content.get('url')
                break
        
        if not servers_zip_url:
            raise Exception("servers.zip not found in Linux target")
        
        # Download and extract servers.zip
        logger.info(f"Downloading servers.zip from {servers_zip_url}")
        zip_response = requests.get(servers_zip_url)
        zip_response.raise_for_status()
        
        # Create temp directory and extract
        temp_dir = tempfile.mkdtemp()
        zip_path = os.path.join(temp_dir, 'servers.zip')
        
        with open(zip_path, 'wb') as f:
            f.write(zip_response.content)
        
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(temp_dir)
        
        # Find aws-lsp-codewhisperer.js
        for root, dirs, files in os.walk(temp_dir):
            if 'aws-lsp-codewhisperer.js' in files:
                LSP_EXECUTABLE_PATH = os.path.join(root, 'aws-lsp-codewhisperer.js')
                logger.info(f"Found LSP executable at: {LSP_EXECUTABLE_PATH}")
                return
        
        raise Exception("aws-lsp-codewhisperer.js not found in extracted files")
        
    except Exception as e:
        logger.error(f"Error downloading LSP server: {e}")
        raise

def _load_jupyter_server_extension(server_app):
    """Registers the API handler to receive HTTP requests from the frontend extension.

    Parameters
    ----------
    server_app: jupyterlab.labapp.LabApp
        JupyterLab application instance
    """
    # TODO remove global variable use 
    global lsp_connection, credential_manager, q_customization
    
    setup_handlers(server_app.web_app)
    
    # Download and extract LSP server
    logger.info("Downloading LSP server")
    download_and_extract_lsp_server()

    # Initialize telemetry logger
    logger.info("Initializing telemetry logger")
    init_api_operation_logger(server_app.log)
    
    # Initialize the LSP connection when the extension loads
    logger.info("Starting LSP server")
    auth_mode = extract_auth_mode()
    lsp_connection = LspServerConnection()
    lsp_connection.start_lsp_server(NODE_PATH, LSP_EXECUTABLE_PATH, auth_mode)
    lsp_connection.initialize(WORKSPACE_FOLDER)
    
    # Initialize credential manager and setup credentials
    credential_manager = CredentialManager(lsp_connection, auth_mode)
    credential_manager.initialize_credentials()
    
    # Update Q profile and customization (only for paid tier)
    from .extract_utils import AuthMode
    if auth_mode == AuthMode.IDC.value:
        # Initialize customization ARN and start file watcher
        customization_arn = extract_q_customization_arn()
        q_customization = QCustomization(lsp_connection)
        q_customization.start_customization_file_watcher()
        q_settings = extract_q_settings()
        if q_settings:
            lsp_connection.update_q_profile(q_settings)
        
        if customization_arn:
            lsp_connection.update_q_customization(customization_arn)
    
    name = "sagemaker_gen_ai_jupyterlab_extension"
    server_app.log.info(f"Registered {name} server extension")

def _unload_jupyter_server_extension(server_app):
    """Unload the extension and stop the LSP server."""
    logger.info("In unload method")
    
    global lsp_connection, credential_manager, q_customization
    try:
        if credential_manager:
            logger.info("Cleaning up credential manager")
            credential_manager.cleanup()
            credential_manager = None
    except Exception as e:
        logger.error(f"Error cleaning up credential manager: {e}")

    # Stop customization file watcher
    try:
        if q_customization:
            logger.info("Stopping Customization file watcher")
            q_customization.stop_watcher_for_customization_file()
    except Exception as e:
         logger.error(f"Error stopping Customization file watcher: {e}")
